/*
 * arch/arm/mach-vexpress/tc2_pm.c - TC2 power management support
 *
 * Created by:	Nicolas Pitre, October 2012
 * Copyright:	(C) 2012  Linaro Limited
 *
 * Some portions of this file were originally written by Achin Gupta
 * Copyright:   (C) 2012  ARM Limited
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/spinlock.h>
#include <linux/errno.h>

#include <asm/bL_entry.h>
#include <asm/proc-fns.h>
#include <asm/cacheflush.h>

#include <mach/motherboard.h>

#include <linux/vexpress.h>
#include <linux/arm-cci.h>

/*
 * We can't use regular spinlocks. In the switcher case, it is possible
 * for an outbound CPU to call power_down() after its inbound counterpart
 * is already live using the same logical CPU number which trips lockdep
 * debugging.
 */
static arch_spinlock_t tc2_pm_lock = __ARCH_SPIN_LOCK_UNLOCKED;

static int tc2_pm_use_count[3][2];

static int tc2_pm_power_up(unsigned int cpu, unsigned int cluster)
{
	pr_debug("%s: cpu %u cluster %u\n", __func__, cpu, cluster);
	if (cluster >= 2 || cpu >= vexpress_spc_get_nb_cpus(cluster))
		return -EINVAL;

	/*
	 * Since this is called with IRQs enabled, and no arch_spin_lock_irq
	 * variant exists, we need to disable IRQs manually here.
	 */
	local_irq_disable();
	arch_spin_lock(&tc2_pm_lock);

	if (!tc2_pm_use_count[0][cluster] &&
	    !tc2_pm_use_count[1][cluster] &&
	    !tc2_pm_use_count[2][cluster])
		vexpress_spc_powerdown_enable(cluster, 0);

	tc2_pm_use_count[cpu][cluster]++;
	if (tc2_pm_use_count[cpu][cluster] == 1) {
		vexpress_spc_write_bxaddr_reg(cluster, cpu,
					      virt_to_phys(bL_entry_point));
		vexpress_spc_set_cpu_wakeup_irq(cpu, cluster, 1);
	} else if (tc2_pm_use_count[cpu][cluster] != 2) {
		/*
		 * The only possible values are:
		 * 0 = CPU down
		 * 1 = CPU (still) up
		 * 2 = CPU requested to be up before it had a chance
		 *     to actually make itself down.
		 * Any other value is a bug.
		 */
		BUG();
	}

	arch_spin_unlock(&tc2_pm_lock);
	local_irq_enable();

	return 0;
}

static void tc2_pm_down(u64 residency)
{
	unsigned int mpidr, cpu, cluster;
	bool last_man = false, skip_wfi = false;

	asm ("mrc p15, 0, %0, c0, c0, 5" : "=r" (mpidr));
	cpu = mpidr & 0xff;
	cluster = (mpidr >> 8) & 0xff;

	pr_debug("%s: cpu %u cluster %u\n", __func__, cpu, cluster);
	BUG_ON(cluster >= 2 || cpu >= vexpress_spc_get_nb_cpus(cluster));

	__bL_cpu_going_down(cpu, cluster);

	arch_spin_lock(&tc2_pm_lock);
	tc2_pm_use_count[cpu][cluster]--;
	if (tc2_pm_use_count[cpu][cluster] == 0) {
		vexpress_spc_set_cpu_wakeup_irq(cpu, cluster, 1);
		if (!tc2_pm_use_count[0][cluster] &&
		    !tc2_pm_use_count[1][cluster] &&
		    !tc2_pm_use_count[2][cluster] &&
		    (!residency || residency > 5000)) {
			vexpress_spc_powerdown_enable(cluster, 1);
			vexpress_spc_set_global_wakeup_intr(1);
			//BUG_ON(__bL_cluster_state(cluster) != CLUSTER_UP);
			if (__bL_cluster_state(cluster) != CLUSTER_UP) {
				pr_crit("%s: last man but __bL_cluster_state(%d) returned %d\n",
					__func__, cluster, __bL_cluster_state(cluster));
				BUG();
			}
			last_man = true;
		}
	} else if (tc2_pm_use_count[cpu][cluster] == 1) {
		/*
		 * A power_up request went ahead of us.
		 * Even if we do not want to shut this CPU down,
		 * the caller expects a certain state as if the WFI
		 * was aborted.  So let's continue with cache cleaning.
		 */
		skip_wfi = true;
	} else
		BUG();

	if (last_man && __bL_outbound_enter_critical(cpu, cluster)) {
		arch_spin_unlock(&tc2_pm_lock);

		asm volatile (
			"mrc	p15, 0, ip, c1, c0, 0 \n\t"
			"bic	ip, ip, #(1 << 2) @ clear C bit \n\t"
			"mcr	p15, 0, ip, c1, c0, 0 \n\t"
			"dsb \n\t"
			"isb"
			: : : "ip" );

		flush_cache_all();

		asm volatile (
			"clrex	\n\t"
			"mrc	p15, 0, ip, c1, c0, 1 \n\t"
			"bic	ip, ip, #(1 << 6) @ clear SMP bit \n\t"
			"mcr	p15, 0, ip, c1, c0, 1 \n\t"
			"isb \n\t"
			"dsb"
			: : : "ip" );

		disable_cci(cluster);

		/*
		 * Ensure that both C & I bits are disabled in the SCTLR
		 * before disabling ACE snoops. This ensures that no
		 * coherency traffic will originate from this cpu after
		 * ACE snoops are turned off.
		 */
		cpu_proc_fin();

		__bL_outbound_leave_critical(cluster, CLUSTER_DOWN);
	} else {

		/*
		 * If last man then undo any setup done previously.
		 */
		if (last_man) {
			vexpress_spc_powerdown_enable(cluster, 0);
			vexpress_spc_set_global_wakeup_intr(0);
		}

		arch_spin_unlock(&tc2_pm_lock);

		asm volatile (
			"mrc	p15, 0, ip, c1, c0, 0 \n\t"
			"bic	ip, ip, #(1 << 2) @ clear C bit \n\t"
			"mcr	p15, 0, ip, c1, c0, 0 \n\t"
			"dsb \n\t"
			"isb"
			: : : "ip" );

		flush_cache_louis();

		asm volatile (
			"clrex	\n\t"
			"mrc	p15, 0, ip, c1, c0, 1 \n\t"
			"bic	ip, ip, #(1 << 6) @ clear SMP bit \n\t"
			"mcr	p15, 0, ip, c1, c0, 1 \n\t"
			"isb \n\t"
			"dsb"
			: : : "ip" );
	}

	__bL_cpu_down(cpu, cluster);

	/* Now we are prepared for power-down, do it: */
	if (!skip_wfi)
		wfi();

	/* Not dead at this point?  Let our caller cope. */
}

static void tc2_pm_power_down(void)
{
	tc2_pm_down(0);
}

static void tc2_pm_suspend(u64 residency)
{
	unsigned int mpidr, cpu, cluster;

	asm ("mrc p15, 0, %0, c0, c0, 5" : "=r" (mpidr));
	cpu = mpidr & 0xff;
	cluster = (mpidr >> 8) & 0xff;
	vexpress_spc_write_bxaddr_reg(cluster, cpu,
				      virt_to_phys(bL_entry_point));
	tc2_pm_down(residency);
}

static void tc2_pm_powered_up(void)
{
	unsigned int mpidr, cpu, cluster;
	unsigned long flags;

	asm ("mrc p15, 0, %0, c0, c0, 5" : "=r" (mpidr));
	cpu = mpidr & 0xff;
	cluster = (mpidr >> 8) & 0xff;

	pr_debug("%s: cpu %u cluster %u\n", __func__, cpu, cluster);
	BUG_ON(cluster >= 2 || cpu >= vexpress_spc_get_nb_cpus(cluster));

	local_irq_save(flags);
	arch_spin_lock(&tc2_pm_lock);

	if (!tc2_pm_use_count[0][cluster] &&
	    !tc2_pm_use_count[1][cluster] &&
	    !tc2_pm_use_count[2][cluster]) {
		vexpress_spc_powerdown_enable(cluster, 0);
		vexpress_spc_set_global_wakeup_intr(0);
	}

	if (!tc2_pm_use_count[cpu][cluster])
		tc2_pm_use_count[cpu][cluster] = 1;

	vexpress_spc_set_cpu_wakeup_irq(cpu, cluster, 0);
	vexpress_spc_write_bxaddr_reg(cluster, cpu, 0);

	arch_spin_unlock(&tc2_pm_lock);
	local_irq_restore(flags);
}

static const struct bL_platform_power_ops tc2_pm_power_ops = {
	.power_up	= tc2_pm_power_up,
	.power_down	= tc2_pm_power_down,
	.suspend	= tc2_pm_suspend,
	.powered_up	= tc2_pm_powered_up,
};

static void __init tc2_pm_usage_count_init(void)
{
	unsigned int mpidr, cpu, cluster;

	asm ("mrc p15, 0, %0, c0, c0, 5" : "=r" (mpidr));
	cpu = mpidr & 0xff;
	cluster = (mpidr >> 8) & 0xff;

	pr_debug("%s: cpu %u cluster %u\n", __func__, cpu, cluster);
	BUG_ON(cpu >= 3 || cluster >= 2);
	tc2_pm_use_count[cpu][cluster] = 1;
}

extern void tc2_pm_power_up_setup(void);

static int __init tc2_pm_init(void)
{
	int ret;

	if (!vexpress_spc_check_loaded())
		return -ENODEV;

	tc2_pm_usage_count_init();

	ret = bL_platform_power_register(&tc2_pm_power_ops);
	if (!ret)
		ret = bL_cluster_sync_init(tc2_pm_power_up_setup);
	if (!ret)
		pr_info("TC2 power management initialized\n");
	return ret;
}

early_initcall(tc2_pm_init);
